Exception: Cohere::Transcribe::Error
- Inherits:
-
StandardError
- Object
- StandardError
- Cohere::Transcribe::Error
- Defined in:
- lib/cohere/transcribe/errors.rb,
sig/cohere/transcribe.rbs
Overview
Keep the generated gem's original base error compatible when this file is loaded directly as well as through the eventual public entry point.